@@ -41,7 +41,7 @@ static float convert_pos (gchar *pos, int digits);
 #endif
 static int compare_country_names (const void *a, const void *b);
 static void sort_locations_by_country (GPtrArray *locations);
-static gchar * tz_data_file_get (void);
+static gchar * tz_data_file_get (gchar *env, gchar *defaultfile);
 
 G_DEFINE_TYPE (CcTimezoneLocation, cc_timezone_location, G_TYPE_OBJECT)
 
@@ -51,7 +51,9 @@ G_DEFINE_TYPE (CcTimezoneLocation, cc_timezone_location, G_TYPE_OBJECT)
 struct _CcTimezoneLocationPrivate
 {
 	gchar *country;
+	gchar *full_country;
 	gchar *en_name;
+	gchar *state;
 	gdouble latitude;
 	gdouble longitude;
 	gchar *zone;
@@ -63,7 +65,9 @@ struct _CcTimezoneLocationPrivate
 enum {
   PROP_0,
   PROP_COUNTRY,
+  PROP_FULL_COUNTRY,
   PROP_EN_NAME,
+  PROP_STATE,
   PROP_LATITUDE,
   PROP_LONGITUDE,
   PROP_ZONE,
@@ -83,9 +87,15 @@ cc_timezone_location_get_property (GObject    *object,
     case PROP_COUNTRY:
       g_value_set_string (value, priv->country);
       break;
+    case PROP_FULL_COUNTRY:
+      g_value_set_string (value, priv->full_country);
+      break;
     case PROP_EN_NAME:
       g_value_set_string (value, priv->en_name);
       break;
+    case PROP_STATE:
+      g_value_set_string (value, priv->state);
+      break;
     case PROP_LATITUDE:
       g_value_set_double (value, priv->latitude);
       break;
@@ -118,9 +128,15 @@ cc_timezone_location_set_property (GObject      *object,
     case PROP_COUNTRY:
       priv->country = g_value_get_string(value);
       break;
+    case PROP_FULL_COUNTRY:
+      priv->full_country = g_value_get_string(value);
+      break;
     case PROP_EN_NAME:
       priv->en_name = g_value_get_string(value);
       break;
+    case PROP_STATE:
+      priv->state = g_value_get_string(value);
+      break;
     case PROP_LATITUDE:
       priv->latitude = g_value_get_double(value);
       break;
@@ -152,6 +168,18 @@ cc_timezone_location_dispose (GObject *object)
       priv->country = NULL;
     }
 
+  if (priv->full_country)
+    {
+      g_free (priv->full_country);
+      priv->full_country = NULL;
+    }
+
+  if (priv->state)
+    {
+      g_free (priv->state);
+      priv->state = NULL;
+    }
+
   if (priv->zone)
     {
       g_free (priv->zone);
@@ -192,12 +220,26 @@ cc_timezone_location_class_init (CcTimezoneLocationClass *klass)
                                           "The country for the location",
                                           "",
                                           G_PARAM_READWRITE));
+  g_object_class_install_property(object_class,
+                                  PROP_FULL_COUNTRY,
+                                  g_param_spec_string ("full_country",
+                                          "Country (full name)",
+                                          "The full country name",
+                                          "",
+                                          G_PARAM_READWRITE));
   g_object_class_install_property(object_class,
                                   PROP_EN_NAME,
                                   g_param_spec_string ("en_name",
                                           "English Name",
                                           "The name of the location",
-					  "",
+                                          "",
+                                          G_PARAM_READWRITE));
+  g_object_class_install_property(object_class,
+                                  PROP_STATE,
+                                  g_param_spec_string ("state",
+                                          "State",
+                                          "The state for the location",
+                                          "",
                                           G_PARAM_READWRITE));
   g_object_class_install_property(object_class,
                                   PROP_LATITUDE,
@@ -254,6 +296,136 @@ cc_timezone_location_new (void)
   return g_object_new (CC_TYPE_TIMEZONE_LOCATION, NULL);
 }
 
+void parse_file (const char * filename,
+                 const guint ncolumns,
+                 GFunc func,
+                 gpointer user_data)
+{
+    FILE *fh = fopen (filename, "r");
+    char buf[4096];
+
+    if (!fh) {
+        g_warning ("Could not open *%s*\n", filename);
+        fclose (fh);
+        return;
+    }
+
+    while (fgets (buf, sizeof(buf), fh))
+    {
+        if (*buf == '#') continue;
+
+        g_strchomp (buf);
+        func (g_strsplit (buf,"\t", ncolumns), user_data);
+    }
+
+    fclose (fh);
+}
+
+void parse_admin1Codes (gpointer parsed_data,
+                        gpointer user_data)
+{
+    gchar ** parsed_data_v = (gchar **) parsed_data;
+    GHashTable * hash_table = (GHashTable *) user_data;
+
+    g_hash_table_insert (hash_table,
+            g_strdup (parsed_data_v[0]),
+            g_strdup (parsed_data_v[1]));
+
+    g_strfreev (parsed_data_v);
+
+}
+
+void parse_countrycode (gpointer parsed_data,
+                        gpointer user_data)
+{
+    gchar ** parsed_data_v = (gchar **) parsed_data;
+    GHashTable * hash_table = (GHashTable *) user_data;
+
+    g_hash_table_insert (hash_table,
+            g_strdup (parsed_data_v[0]),
+            g_strdup (parsed_data_v[4]));
+
+    g_strfreev (parsed_data_v);
+}
+
+typedef struct Triple {
+    gpointer first;
+    gpointer second;
+    gpointer third;
+} Triple;
+
+void parse_cities15000 (gpointer parsed_data, 
+                        gpointer user_data)
+{
+    gchar ** parsed_data_v = (gchar **) parsed_data;
+    Triple * triple = (Triple *) user_data;
+    GPtrArray * ptr_array = (GPtrArray *) triple->first;
+    GHashTable * stateHash = (GHashTable *) triple->second;
+    GHashTable * countryHash = (GHashTable *) triple->third;
+
+    CcTimezoneLocation *loc = cc_timezone_location_new ();
+
+    loc->priv->country = g_strdup (parsed_data_v[8]);
+    loc->priv->en_name = g_strdup (parsed_data_v[2]);
+
+    gchar * tmpState = g_strdup_printf ("%s.%s", loc->priv->country,
+            parsed_data_v[10]);
+    loc->priv->state = g_strdup (
+            (gchar *) g_hash_table_lookup (
+                stateHash,
+                tmpState));
+    g_free (tmpState);
+
+    loc->priv->full_country = g_strdup (
+            (gchar *) g_hash_table_lookup (
+                countryHash,
+                loc->priv->country));
+
+    loc->priv->zone = g_strdup (parsed_data_v[17]);
+    loc->priv->latitude  = g_ascii_strtod(parsed_data_v[4], NULL);
+    loc->priv->longitude = g_ascii_strtod(parsed_data_v[5], NULL);
+
+#ifdef __sun
+    gchar *latstr, *lngstr, *p;
+
+    latstr = g_strdup (parsed_data_v[1]);
+    p = latstr + 1;
+    while (*p != '-' && *p != '+') p++;
+    lngstr = g_strdup (p);
+    *p = '\0';
+
+    if (parsed_data_v[3] && *parsed_data_v[3] == '-' && parsed_data_v[4])
+        loc->comment = g_strdup (parsed_data_v[4]);
+
+    if (parsed_data_v[3] && *parsed_data_v[3] != '-' && !islower(loc->zone)) {
+        CcTimezoneLocation *locgrp;
+
+        /* duplicate entry */
+        locgrp = cc_timezone_location_new ();
+        locgrp->country = g_strdup (parsed_data_v[0]);
+        locgrp->en_name = NULL;
+        locgrp->zone = g_strdup (parsed_data_v[3]);
+        locgrp->latitude  = convert_pos (latstr, 2);
+        locgrp->longitude = convert_pos (lngstr, 3);
+        locgrp->comment = (parsed_data_v[4]) ? g_strdup (parsed_data_v[4]) : NULL;
+
+        g_ptr_array_add (ptr_array, (gpointer) locgrp);
+    }
+#else
+    loc->priv->comment = NULL;
+#endif
+
+    g_ptr_array_add (ptr_array, (gpointer) loc);
+
+#ifdef __sun
+    g_free (latstr);
+    g_free (lngstr);
+#endif
+    g_strfreev (parsed_data_v);
+
+    return;
+}
+
 
 /* ---------------- *
  * Public interface *
@@ -261,91 +433,60 @@ cc_timezone_location_new (void)
 TzDB *
 tz_load_db (void)
 {
-	gchar *tz_data_file;
-	TzDB *tz_db;
-	FILE *tzfile;
-	char buf[4096];
+    gchar *tz_data_file, *admin1_file, *country_file;
+    TzDB *tz_db;
+    char buf[4096];
 
-	tz_data_file = tz_data_file_get ();
-	if (!tz_data_file) {
-		g_warning ("Could not get the TimeZone data file name");
-		return NULL;
-	}
-	tzfile = fopen (tz_data_file, "r");
-	if (!tzfile) {
-		g_warning ("Could not open *%s*\n", tz_data_file);
-		g_free (tz_data_file);
-		return NULL;
-	}
+    tz_data_file = tz_data_file_get ("TZ_DATA_FILE", TZ_DATA_FILE);
+    if (!tz_data_file) {
+        g_warning ("Could not get the TimeZone data file name");
+        return NULL;
+    }
 
-	tz_db = g_new0 (TzDB, 1);
-	tz_db->locations = g_ptr_array_new ();
+    admin1_file = tz_data_file_get ("ADMIN1_FILE", ADMIN1_FILE);
+    if (!admin1_file) {
+        g_warning ("Could not get the admin1 data file name");
+        return NULL;
+    }
 
-	while (fgets (buf, sizeof(buf), tzfile))
-	{
-		gchar **tmpstrarr;
-		CcTimezoneLocation *loc;
+    country_file = tz_data_file_get ("COUNTRY_FILE", COUNTRY_FILE);
+    if (!country_file) {
+        g_warning ("Could not get the country data file name");
+        return NULL;
+    }
 
-		if (*buf == '#') continue;
+    GHashTable *stateHash = g_hash_table_new_full (g_str_hash,
+            g_str_equal, g_free, g_free);
 
-		g_strchomp(buf);
-		tmpstrarr = g_strsplit(buf,"\t", 19);
-		
-		loc = cc_timezone_location_new ();
-		loc->priv->country = g_strdup (tmpstrarr[8]);
-		loc->priv->en_name = g_strdup (tmpstrarr[2]);
-		loc->priv->zone = g_strdup (tmpstrarr[17]);
-		loc->priv->latitude  = g_ascii_strtod(tmpstrarr[4], NULL);
-		loc->priv->longitude = g_ascii_strtod(tmpstrarr[5], NULL);
-		
-#ifdef __sun
-		gchar *latstr, *lngstr, *p;
+    parse_file (admin1_file, 4, parse_admin1Codes, stateHash);
 
-		latstr = g_strdup (tmpstrarr[1]);
-		p = latstr + 1;
-		while (*p != '-' && *p != '+') p++;
-		lngstr = g_strdup (p);
-		*p = '\0';
-		
-		if (tmpstrarr[3] && *tmpstrarr[3] == '-' && tmpstrarr[4])
-			loc->comment = g_strdup (tmpstrarr[4]);
+    GHashTable * countryHash = g_hash_table_new_full (g_str_hash,
+            g_str_equal, g_free, g_free);
 
-		if (tmpstrarr[3] && *tmpstrarr[3] != '-' && !islower(loc->zone)) {
-			CcTimezoneLocation *locgrp;
+    parse_file (country_file, 19, parse_countrycode, countryHash);
 
-			/* duplicate entry */
-			locgrp = cc_timezone_location_new ();
-			locgrp->country = g_strdup (tmpstrarr[0]);
-			locgrp->en_name = NULL;
-			locgrp->zone = g_strdup (tmpstrarr[3]);
-			locgrp->latitude  = convert_pos (latstr, 2);
-			locgrp->longitude = convert_pos (lngstr, 3);
-			locgrp->comment = (tmpstrarr[4]) ? g_strdup (tmpstrarr[4]) : NULL;
+    tz_db = g_new0 (TzDB, 1);
+    tz_db->locations = g_ptr_array_new ();
 
-			g_ptr_array_add (tz_db->locations, (gpointer) locgrp);
-		}
-#else
-		loc->priv->comment = NULL;
-#endif
+    Triple * triple = g_new (Triple, 1);
+    triple->first = tz_db->locations;
+    triple->second = stateHash;
+    triple->third = countryHash;
 
-		g_ptr_array_add (tz_db->locations, (gpointer) loc);
+    parse_file (tz_data_file, 19, parse_cities15000, triple);
 
-#ifdef __sun
-		g_free (latstr);
-		g_free (lngstr);
-#endif
-		g_strfreev (tmpstrarr);
-	}
-	
-	fclose (tzfile);
-	
-	/* now sort by country */
-	sort_locations_by_country (tz_db->locations);
-	
-	g_free (tz_data_file);
-	
-	return tz_db;
-}    
+    g_hash_table_destroy (stateHash);
+    g_hash_table_destroy (countryHash);
+    triple->second = NULL;
+    triple->third = NULL;
+
+    /* now sort by country */
+    sort_locations_by_country (tz_db->locations);
+
+    g_free (tz_data_file);
+
+    return tz_db;
+}
 
 void
 tz_db_free (TzDB *db)
@@ -511,13 +652,13 @@ tz_info_free (TzInfo *tzinfo)
  * ----------------- */
 
 static gchar *
-tz_data_file_get (void)
+tz_data_file_get (gchar *env, gchar *defaultfile)
 {
-	gchar *file;
+    /* Allow passing this in at runtime, to support loading it from the build
+     * tree during tests. */
+    const gchar * filename = g_getenv (env);
 
-	file = g_strdup (TZ_DATA_FILE);
-
-	return file;
+    return filename ? g_strdup (filename) : g_strdup (defaultfile);
 }
 
 #ifdef __sun
